abstract type AbstractBoxing end

struct NoBoxing <: AbstractBoxing end
NoBoxing(::Simulation) = NoBoxing()

"""
    BoxGrid(p, w)
Internal container to pass around precomputed boxes
"""
struct BoxGrid{N, T, P} <: AbstractBoxing
    size::NTuple{N,T}
    boxes::Array{Vector{P}, N}
    neighborlist::Array{Vector{CartesianIndex{N}}, N}

    function BoxGrid{N, T, P}(sim::Simulation) where {N, P,T}
        nboxes = Tuple(Int.(domainsize(sim) .÷ interactionrange(P, sim)))
        trivial = false
        if any(nboxes .== 0)
            @warn "Domain smaller than interaction range, initializing trivial boxgrid"
            nboxes = Tuple(fill(1, length(domainsize(sim))))
            trivial = true
        end
        lengths = Tuple(domainsize(sim) ./ nboxes)
        boxes = [ P[] for i in CartesianIndices(nboxes)]

        neighborlist = map(CartesianIndices(boxes)) do idx
            nidx = trivial ? Vector{CartesianIndex{N}}() : compute_neighboring_boxes(idx, sim.domain, nboxes)
            if !(all(nidx .∈ Ref(CartesianIndices(boxes))))
                @warn "$idx has invalid neighbor $nidx"
            end
            nidx
        end
        new{N, T, P}(lengths, boxes, neighborlist)
    end
end

BoxGrid(sim::Simulation{PTC}, T::Type, P::Type) where PTC = BoxGrid{ndims(PTC), T, P}(sim)

"""
    boxparticles!(bg::BoxGrid, partices)
Sorts the particles into boxes of the given `BoxGrid`.
"""
function boxparticles!(bg::BoxGrid{N}, particles) where {N}

    # Clear boxes before refilling
    for box in bg.boxes
        empty!(box)
    end

    # TODO: test whether n-D version is slower
    ci = CartesianIndices(bg.boxes)
    @inbounds for c in particles
        actual_index = CartesianIndex{N}(
            # HACK: this depends on (::SArray).data being a tuple
            @.(unsafe_trunc(Int, c.centerpos.data / (bg.size + 1e-12 ) + 1))
        )
        if actual_index in ci
            box = bg.boxes[actual_index]
            push!(box, c)
        else
            error("No suitable box found for a particle with position $(c.centerpos)")
        end
    end
    return nothing
end



"relative positions of upper right-hand neighbors on a 2D grid"
const relneighbors2D = (
    CartesianIndex( 0, 1),
    CartesianIndex( 1,-1),
    CartesianIndex( 1, 0),
    CartesianIndex( 1, 1),
)

"relative positions of upper right-hand neighbors on a 3D grid"
const relneighbors3D = (
    CartesianIndex( 0, 1, 0),
    CartesianIndex( 0, 0, 1),
    CartesianIndex( 1, 0, 0),
    CartesianIndex( 1, 0, 1),
    CartesianIndex( 1, 1, 0),
    CartesianIndex( 0, 1, 1),
    CartesianIndex( 1, 1, 1),
    CartesianIndex( 1, 0,-1),
    CartesianIndex( 1,-1, 0),
    CartesianIndex( 0,-1, 1),
    CartesianIndex( 1,-1, 1),
    CartesianIndex( 1, 1,-1),
    CartesianIndex( 1,-1,-1),
)


"""
    compute_neighboring_boxes(idx, d::AbstractDomain, nboxes)
Finds all upper right-hand neighbors of the given box.
"""
function compute_neighboring_boxes(idx, ::Domain2D{B1, B2}, nboxes) where {B1,B2}
    nlist = Set{CartesianIndex{2}}()
    for offs in relneighbors2D
        neighbor = idx + offs
        i, j = neighbor.I
        if B1 <: PeriodicBoundary
            i = mod1(i, nboxes[1])
        end
        if B2 <: PeriodicBoundary
            j = mod1(j, nboxes[2])
        end
        neighbor = CartesianIndex(i,j)
        if neighbor != idx && neighbor ∈ CartesianIndices(nboxes)
            push!(nlist, neighbor)
        end
    end
    collect(nlist)
end

function compute_neighboring_boxes(idx, ::Domain3D{B1, B2, B3}, nboxes) where {B1,B2, B3}
    nlist = Set{CartesianIndex{3}}()
    for offs in relneighbors3D
        neighbor = idx + offs
        i, j, k = neighbor.I
        if B1 <: PeriodicBoundary
            i = mod1(i, nboxes[1])
        end
        if B2 <: PeriodicBoundary
            j = mod1(j, nboxes[2])
        end
        if B3 <: PeriodicBoundary
            k = mod1(k, nboxes[3])
        end
        neighbor = CartesianIndex(i,j, k)
        if neighbor != idx && neighbor ∈ CartesianIndices(nboxes)
            push!(nlist, neighbor)
        end
    end
    collect(nlist)
end

"""
    relevant_boxes(so::AbstractObstacle, bg::BoxGrid, d::AbstractDomain)
Compute the boxes of `bg` that contains points less then one interaction radius away from so.

Returns a vector of indices.
"""
function relevant_boxes(so::AbstractObstacle, bg::BoxGrid, w::AbstractDomain{2}; kwargs...)
    #the interactionrange by which the polygon should be grown is mininum(bg.size)/2
    ip = interactionpolygon(so, w; growby=minimum(bg.size)/2, kwargs...)

    boxmat = dropdims(intersecting_boxes(ip, w, size(bg.boxes)), dims=3)
    vec(CartesianIndices(bg.boxes))[vec(boxmat) .> 0]
end


function relevant_boxes(so::AbstractObstacle, bg::BoxGrid, w::AbstractDomain{3}; kwargs...)
    # HACK: this is not how it works.
    # well it is now.
    @warn "Automatic box detection for 3D obstacles is not supported. Override `relevant_boxes` for your obstacle to manually select boxes" maxlog=1
    return vec(CartesianIndices(bg.boxes))
end



## BoxGrid pairwise interaction computation

@forcedef pairwise_interactions(interaction::Function, sim::Simulation, bg::BoxGrid) =
    @forcefun pairwise_interactions(interaction, sim, bg, Val{isthreaded(sim)}())


"""
    pairwise_interactions(interaction::Function, sim::Simulation, bg::BoxGrid, ::Val{false})
Uses a [`BoxGrid`](@ref) to efficiently compute pairwise interactions.
`BoxGrid` structure is mutated in the sense that particles are sorted into their boxes.
"""
@forcedef function pairwise_interactions(interaction::F, sim::Simulation, bg::BoxGrid, ::Val{false}) where {F<:Function}
    # Type parameter F is not needed within the function but apparently causes the compiler to
    # to properly infer the type of the interaction function and hence not use runtime dispatch

    # For execution order preservation compared to multithreaded simulations the
    # boxes are iterated in the same striped manner.
    boxparticles!(bg, particles(sim))

    numstripes, numrows... = size(bg.boxes)
    rowit = CartesianIndices(numrows)

    for i = 1:2:numstripes
        for j = rowit
            I = CartesianIndex(i,j)
            isempty(bg.boxes[I]) || @forcefun pairwise_interactions(interaction, sim, bg.boxes[I])
            for J ∈ bg.neighborlist[I]
                if !isempty(bg.boxes[J])
                    @forcefun pairwise_interactions(interaction, sim, bg.boxes[I], bg.boxes[J])
                end
            end # for J ∈ neighbors
        end
    end
    for i = 2:2:numstripes
        for j = rowit
            I = CartesianIndex(i,j)
            isempty(bg.boxes[I]) || @forcefun pairwise_interactions(interaction, sim, bg.boxes[I])
            for J ∈ bg.neighborlist[I]
                if !isempty(bg.boxes[J])
                    @forcefun pairwise_interactions(interaction, sim, bg.boxes[I], bg.boxes[J])
                end
            end # for J ∈ neighbors
        end
    end
    nothing
end


"""
    pairwise_interactions(interaction::Function, sim::Simulation, bg::BoxGrid, ::Val{true})
Multi-threaded version of [`pairwise_interactions!_serial`](@ref).

## Extended help

Multithreading is implemented by dividing the system into two sets of interleaved stripes (slabs in 3D)
of boxes, each stripe being one box wide.
This division ensures that each stripe is independent of all other stripes in same set.

Therefore, we can first integrate all the stripes in the first set in parallel, synchronise, then
integrate the second set.
"""
@forcedef function pairwise_interactions(interaction::F, sim::Simulation, bg::BoxGrid, ::Val{true}) where {F<:Function}
    boxparticles!(bg, particles(sim))

    numstripes, numrows... = size(bg.boxes)
    rowit = CartesianIndices(numrows)

    Threads.@threads for i = 1:2:numstripes
        for j = rowit
            I = CartesianIndex(i,j)
            isempty(bg.boxes[I]) || @forcefun pairwise_interactions(interaction, sim, bg.boxes[I])
            for J ∈ bg.neighborlist[I]
                if !isempty(bg.boxes[J])
                    @forcefun pairwise_interactions(interaction, sim, bg.boxes[I], bg.boxes[J])
                end
            end # for J ∈ neighbors
        end
    end
    Threads.@threads for i = 2:2:numstripes
        for j = rowit
            I = CartesianIndex(i,j)
            isempty(bg.boxes[I]) || @forcefun pairwise_interactions(interaction, sim, bg.boxes[I])
            for J ∈ bg.neighborlist[I]
                if !isempty(bg.boxes[J])
                    @forcefun pairwise_interactions(interaction, sim, bg.boxes[I], bg.boxes[J])
                end
            end # for J ∈ neighbors
        end
    end
    nothing
end


@forcedef function pairwise_interactions(interaction::Function, sim::Simulation, ::NoBoxing)
    @forcefun pairwise_interactions(interaction, sim, particles(sim))
end

@forcedef function pairwise_interactions(interaction::Function, sim::Simulation,
                                         particles::AbstractVector{<: AbstractParticle})
    N = length(particles)
    for i = 1:N
        c1 = particles[i]
        for j = i+1:N
            @forcefun interaction(c1, particles[j], sim)
        end
    end
    nothing
end

@forcedef function pairwise_interactions(interaction::Function, sim::Simulation,
                                         particles1::AbstractVector{P},
                                         particles2::AbstractVector{P}) where {P <: AbstractParticle}
    for i in eachindex(particles1)
        c1 = particles1[i]
        for j in eachindex(particles2)
            @forcefun interaction(c1, particles2[j], sim)
        end
    end
    nothing
end
